import ast
import subprocess
from pathlib import Path
from genpydoc.config.config import Config

DocumentableFunc = ast.AsyncFunctionDef | ast.FunctionDef
DocumentableFuncOrClass = DocumentableFunc | ast.ClassDef
DocumentableNode = DocumentableFuncOrClass | ast.Module


class Transformer(ast.NodeTransformer):
    config: Config
    comments: dict[str, str]

    def __init__(self, config: Config, comments: dict[str, str]):
        super().__init__()
        self.config = config
        self.comments = comments

    def _visit_helper(self, node: DocumentableNode) -> DocumentableNode:
        if node.name not in self.comments:
            return node
        new_docstring = (
            self.comments[node.name].removeprefix('"""').removesuffix('"""')
        )
        new_doc_code = ast.Expr(value=ast.Constant(value=new_docstring))
        if (
            node.body
            and isinstance(node.body[0], ast.Expr)
            and isinstance(getattr(node.body[0], "value", None), ast.Constant)
            and isinstance(node.body[0].value.value, str)
        ):
            node.body[0] = new_doc_code
        else:
            node.body.insert(0, new_doc_code)
        return node

    def visit_ClassDef(
        self, node: DocumentableFuncOrClass
    ) -> DocumentableFuncOrClass:
        return self._visit_helper(node=node)

    def visit_FunctionDef(
        self, node: DocumentableFuncOrClass
    ) -> DocumentableFuncOrClass:
        return self._visit_helper(node=node)

    def visit_AsyncFunctionDef(
        self, node: DocumentableFuncOrClass
    ) -> DocumentableFuncOrClass:
        return self._visit_helper(node=node)


class Parser:
    def __init__(self, config: Config):
        self.config = config

    def process(self, filepath: Path, comments: dict[str, str]) -> None:
        with open(filepath) as file:
            source = file.read()
        parsed_tree = ast.parse(source)
        t = Transformer(config=self.config, comments=comments)
        t.visit(parsed_tree)
        filepath.write_text(ast.unparse(parsed_tree))
        if (
            self.config.post_processing.cleanup
            or self.config.post_processing.convert
        ):
            self.post_process(
                filepath,
                cleanup=self.config.post_processing.cleanup,
                convert=self.config.post_processing.convert,
            )

    def post_process(
        self, filepath: str | Path, cleanup: bool = True, convert: bool = True
    ) -> None:
        if cleanup:
            subprocess.run(f'black "{filepath}" -q', shell=True)
        if convert:
            if self.config.docstring_style not in [
                "google",
                "numpy",
                "epytext",
                "reST",
            ]:
                raise ValueError(
                    "Style cannot be converted to with docconvert."
                )
            subprocess.run(
                f'docconvert "{filepath}" --output {self.config.docstring_style} --in-place',
                shell=True,
                input="y\n",
                text=True,
            )



import re
from dataclasses import dataclass
from enum import StrEnum
from typing import Optional

from git import Diff

HUNK_REGEX = re.compile(r"^@@ -(\d+)(?:,(\d+))? \+(\d+)(?:,(\d+))? @@")


class DiffChangeType(StrEnum):
    # '+', '-', ' '
    ADD = "+"
    REMOVE = "-"
    BLANK = " "


@dataclass
class DiffChange:
    old_lineno: Optional[int]
    new_lineno: Optional[int]
    kind: str
    text: str


class ChangeType(StrEnum):
    ADDED = "A"
    DELETED = "D"
    # C = "C"
    MODIFIED = "M"
    RENAMED = "R"
    TYPE_PATH = "T"
    # U = "U"


def parse_diff(diff_text: str | bytes | None) -> list[DiffChange]:
    changes = []

    old_lineno = None
    new_lineno = None

    if isinstance(diff_text, bytes):
        diff_text = diff_text.decode("utf-8")

    for line in diff_text.splitlines():
        m = HUNK_REGEX.match(line)
        if m:
            old_lineno = int(m.group(1))
            new_lineno = int(m.group(3))
            continue

        if line.startswith("---") or line.startswith("+++"):
            # skip file headers
            continue

        if old_lineno is None or new_lineno is None:
            # not inside a hunk yet
            continue

        prefix = line[0]
        text = line[1:] if len(line) else ""

        if prefix == " ":
            changes.append(
                DiffChange(old_lineno, new_lineno, DiffChangeType.BLANK, text)
            )
            old_lineno += 1
            new_lineno += 1
        elif prefix == "+":
            changes.append(
                DiffChange(None, new_lineno, DiffChangeType.ADD, text)
            )
            new_lineno += 1
        elif prefix == "-":
            changes.append(
                DiffChange(old_lineno, None, DiffChangeType.REMOVE, text)
            )
            old_lineno += 1
        else:
            # e.g. "\ No newline at end of file"
            pass
    return changes


def process_changes(changes: list[DiffChange]) -> set[int]:
    lines = set()
    for change in changes:
        if change.kind == DiffChangeType.BLANK:
            continue
        if change.text.strip() == "":
            # whitespace or blank line
            continue

        lineno = next(
            (
                item
                for item in [change.new_lineno, change.old_lineno]
                if item is not None
            ),
            -1,
        )
        if lineno == -1:
            raise
        lines.add(lineno)
    return lines


def process_git_diff(diff: Diff) -> set[int]:
    return process_changes(parse_diff(diff.diff))


def get_change_type(diff: Diff) -> ChangeType:
    if diff.change_type:
        return ChangeType(diff.change_type)
    elif diff.a_path is None and diff.b_path is not None:
        return ChangeType.DELETED
    elif diff.a_path is not None and diff.b_path is None:
        return ChangeType.ADDED
    elif (
        diff.a_path is not None
        and diff.b_path is not None
        and diff.a_path != diff.b_path
    ):
        return ChangeType.RENAMED
    elif (
        diff.a_path is not None
        and diff.b_path is not None
        and diff.a_path == diff.b_path
    ):
        return ChangeType.MODIFIED
    else:
        raise ValueError(diff)



import asyncio
import os
from pathlib import Path
from timeit import default_timer as timer

from dotenv import load_dotenv
from openai import AsyncOpenAI as AsyncOpenAIClient
from tqdm import tqdm

from genpydoc.commenter.transformer import Parser
from genpydoc.config.config import Config
from genpydoc.extractor.visit import CovNode

load_dotenv()

OPENAI_API_KEY = os.getenv("OPENAI_API_KEY")


class Commenter:
    MODELS_MAPPING: dict = {
        "openai": "gpt-5-nano",
    }

    _model: str = None
    _client_api: AsyncOpenAIClient

    def __init__(self, config: Config):
        self.config = config
        self.__init_client()
        self.parser = Parser(config=config)

    def __init_client(self) -> None:
        self._model = self.config.use_model
        if self.config.use_llm_provider == "openai":  # todo other api?
            if not OPENAI_API_KEY:
                raise EnvironmentError("No API key.")
            self._client_api = AsyncOpenAIClient(api_key=OPENAI_API_KEY)

    def build_prompt(self, node: CovNode) -> str:
        if node.node_type == "ClassDef":
            return self.build_class_prompt(node)
        elif node.node_type in ["FunctionDef", "AsyncFunctionDef"]:
            return self.build_function_prompt(node)
        return ""

    def build_class_prompt(self, node: CovNode) -> str:
        return (
            f"You are a senior python engineer. Analyze the following class between <code.start> and <code.end>:\n\n<code.start>\n{node.code}\n<code.end>\n\n"
            + (
                f"This is the current docstring between <doc.start> and <doc.end>:\n\n<doc.start>\n{node.docstring}\n<doc.end>\n\n"
                if node.docstring
                else ""
            )
            + f"I want you to analyze the purpose of the class, and write a new docstring for {node.name} (and only for {node.name}).\n"
            f"The docstring must be using {self.config.docstring_style} style. Since this is a class, only write a description for the purpose of the class, and list the attributes. "
            "If the old docstring correctly reflects the purpose of the code segment, return -1, else return only the docstring."
        )

    def build_function_prompt(self, node: CovNode) -> str:
        return (
            f"You are a senior python engineer. Analyze the following code block between <code.start> and <code.end>:\n\n<code.start>\n{node.code}\n<code.end>\n\n"
            + (
                f"This is the current docstring between <doc.start> and <doc.end>:\n\n<doc.start>\n{node.docstring}\n<doc.end>\n\n"
                if node.docstring
                else ""
            )
            + f"I want you to analyze the purpose of the code segment, and write a new docstring for {node.name} (and only for {node.name}).\n"
            f"The docstring must be using {self.config.docstring_style} style. You must only write a description of the function, "
            f"list the attributes with a basic description, explain any errors raised. Only explicitly show a return if the function returns something not None. "
            "If the old docstring correctly reflects the purpose of the code segment, return -1, else return only the docstring."
        )

    async def openai_process(
        self, prompt: str, node_name: str, responses: dict[str, str]
    ) -> None:
        response = await self._client_api.responses.create(
            model=self._model, input=prompt
        )
        responses[node_name] = response.output_text

    async def process_prompt(
        self, prompt: str, node_name: str, responses: dict[str, str]
    ) -> None:
        if self.config.use_llm_provider == "openai":
            await self.openai_process(prompt, node_name, responses)

    async def process_prompts(self, prompts: dict[str, str]) -> dict[str, str]:
        responses = {}
        start = timer()

        tasks = [
            asyncio.create_task(self.process_prompt(prompt, name, responses))
            for name, prompt in prompts.items()
        ]

        with tqdm(
            total=len(tasks), desc="Commenting", unit="block", leave=True
        ) as pbar:
            for fut in asyncio.as_completed(tasks):
                try:
                    await fut
                except Exception as e:
                    tqdm.write(f"[error]: a task failed: {e!r}")
                finally:
                    pbar.update(1)

        end = timer()
        print(f"Generated comments concurrently in {end - start:.2f} seconds.")
        return responses

    async def comment(self, nodes: list[CovNode]) -> dict[str, str]:
        not_ignored = [
            n
            for n in nodes
            if n.node_type in ["ClassDef", "FunctionDef", "AsyncFunctionDef"]
        ]
        prompts = {
            node.name: self.build_prompt(node=node) for node in not_ignored
        }
        comments = await self.process_prompts(prompts=prompts)
        return comments

    def document(self, nodes: dict[str, list[CovNode]]) -> None:
        for file in nodes:
            print(f"Checking file: {file}.", end="\t")
            docs = asyncio.run(self.comment(nodes[file]))
            docs = {k: v for k, v in docs.items() if v != "-1"}
            self.parser.process(Path(file), docs)



import tomllib
from typing import Any, Literal

import attr


class PostProcessingConfig:
    cleanup: bool = attr.ib(default=True)
    convert: bool = attr.ib(default=True)


@attr.s
class Config:
    VALID_STYLES = ("sphinx", "google")  # FIXME needed?
    VALID_LLM_PROVIDERS = ("openai",)

    docstring_style: str = attr.ib(default="sphinx")
    ignore_magic: bool = attr.ib(default=False)
    ignore_module: bool = attr.ib(default=True)
    ignore_private: bool = attr.ib(default=False)
    ignore_semiprivate: bool = attr.ib(default=False)
    ignore_init_method: bool = attr.ib(default=True)
    ignore_nested_classes: bool = attr.ib(default=False)
    ignore_nested_functions: bool = attr.ib(default=False)
    ignore_property_setters: bool = attr.ib(default=False)
    ignore_property_decorators: bool = attr.ib(default=False)
    ignore_overloaded_functions: bool = attr.ib(default=False)
    include_only_covered: bool = attr.ib(default=True)
    run_on_diff: bool = attr.ib(default=True)
    use_llm_provider: Literal["openai"] = attr.ib(
        default="openai"
    )  # TODO validate
    use_model: str = attr.ib(default="gpt-5-nano")  # TODO validate
    post_processing: PostProcessingConfig = attr.ib(
        default=PostProcessingConfig()
    )

    @use_llm_provider.validator
    def _validate_llm_provider(self, _attribute, value) -> None:
        if value not in self.VALID_LLM_PROVIDERS:
            raise ValueError(
                f"Invalid LLM provider '{value}'.\nSelect one of the following: {', '.join(self.VALID_LLM_PROVIDERS)}"
            )

    @docstring_style.validator
    def _validate_style(self, _attribute, value) -> None:
        if value not in self.VALID_STYLES:
            raise ValueError(
                f"invalid docstring_style '{value}'.\nSelect one of the following: {', '.join(self.VALID_STYLES)}"
            )


def parse_pyproject_toml(path_config: str) -> dict[str, Any] | None:
    with open(path_config, "rb") as file:
        toml = tomllib.load(file)
    config = toml.get("tool", {}).get("genpydoc")
    return {k.replace("-", "_"): v for k, v in config.items()}



import os

from genpydoc.commenter import Commenter
from genpydoc.config.config import Config
from genpydoc.utils.utils import find_project_root
from genpydoc.extractor.extract import Extract
from genpydoc.git_retriever.git_retriever import GitRetriever


class Documenter:
    config: Config

    def __init__(self, config: Config):
        self.config = config

    def document(self, paths):
        extract = Extract(paths, self.config)
        all_nodes, covered_nodes = extract.get_coverage()

        if self.config.include_only_covered:
            nodes = covered_nodes
        else:
            nodes = all_nodes

        if self.config.run_on_diff:
            print("Filtering on git changes...")
            root = find_project_root((os.path.dirname(__file__),))
            gitter = GitRetriever(root, covered_nodes, nodes)
            nodes = gitter.extract_diff()

        if nodes:
            commenter = Commenter(config=self.config)
            commenter.document(nodes=nodes)
        else:
            print("Nothing to comment.")



import ast
import os.path
import pathlib
import sys
from fnmatch import fnmatch
from pathlib import Path
from typing import Iterator

from genpydoc.config.config import Config
from genpydoc.extractor.visit import CovNode
from genpydoc.utils.utils import get_common_base
from genpydoc.extractor import visit


class Extract:
    COMMON_EXCLUDE = [".tox", ".venv", "venv", ".git", ".hg"]
    VALID_EXTENSIONS = [".py", ".pyi"]

    def __init__(self, paths, config: Config | None = None):
        self.paths = paths
        self.extensions = set(".py")
        self.config = config if config else Config()
        self.excluded = ()
        self.common_base = pathlib.Path("/")
        self.output_formatter = None
        self._add_common_exclude()
        self.skipped_file_count = 0

    def _add_common_exclude(self) -> None:
        for path in self.paths:
            self.excluded = self.excluded + tuple(
                os.path.join(path, i) for i in self.COMMON_EXCLUDE
            )

    def _filter_files(self, files: list[str]) -> Iterator[str]:
        for file in files:
            has_valid_ext = any(
                [file.endswith(ext) for ext in self.extensions]
            )
            if not has_valid_ext:
                continue
            basename = os.path.basename(file)
            if basename == "__init__":  # always ignore __init__ files
                continue
            if any(fnmatch(file, exc + "*") for exc in self.excluded):
                continue
            yield file

    def _filter_nodes(self, nodes: list[CovNode]) -> list[CovNode]:
        if self.config.ignore_module:
            return [node for node in nodes if node.node_type != "Module"]
        return nodes

    @staticmethod
    def _filter_empty_nodes(
        nodes: list[CovNode] | None,
    ) -> list[CovNode] | None:
        if not nodes:
            return None
        return [node for node in nodes if node.covered]

    @staticmethod
    def _filter_inner_nested(nodes: list[CovNode]) -> list[CovNode]:
        nested_cls = [n for n in nodes if n.is_nested_cls]
        inner_nested_nodes = [n for n in nodes if n.parent in nested_cls]
        filtered_nodes = [n for n in nodes if n not in inner_nested_nodes]
        filtered_nodes = [n for n in filtered_nodes if n not in nested_cls]
        return filtered_nodes

    @staticmethod
    def _set_google_style(nodes: list[CovNode]) -> None:
        for node in nodes:
            if node.node_type == "FunctionDef" and node.name == "__init__":
                if not node.covered and node.parent.covered:
                    setattr(node, "covered", True)
                elif node.covered and not node.parent.covered:
                    setattr(node.parent, "covered", True)

    def get_filenames_from_path(self) -> list[str]:
        filenames = []
        for path in self.paths:
            if os.path.isfile(path):
                has_valid_extension = any(
                    path.endswith(ext) for ext in self.VALID_EXTENSIONS
                )
                if not has_valid_extension:
                    print(f"invalid file {path}")
                    return sys.exit(1)
                filenames.append(path)
                continue

            for root, _, fs in os.walk(path):
                full_paths = [os.path.join(root, f) for f in fs]
                filenames.extend(self._filter_files(full_paths))

        if not filenames:
            p = ", ".join(self.paths)
            print(f"no python files found in {p}")
            return sys.exit(1)

        self.common_base = get_common_base(filenames)
        return filenames

    def _get_coverage(
        self, filenames: list[str | Path]
    ) -> tuple[dict[str, list[CovNode]], dict[str, list[CovNode]]]:
        results: dict[str, list[CovNode]] = {}
        covered_results: dict[str, list[CovNode]] = {}
        for filename in filenames:
            result = self._get_file_coverage(filename)
            covered_result = self._filter_empty_nodes(result)
            if result:
                results[filename] = result
            if covered_result:
                covered_results[filename] = covered_result
        return results, covered_results

    def _get_file_coverage(self, filename: str | Path) -> list[CovNode] | None:
        with open(filename) as f:
            source = f.read()

        parsed_tree = ast.parse(source)
        visitor = visit.Visitor(filename, self.config, source)
        visitor.visit(parsed_tree)

        filtered_nodes = self._filter_nodes(visitor.nodes)
        if len(filtered_nodes) == 0:
            return None

        if self.config.ignore_nested_functions:
            filtered_nodes = [
                node for node in filtered_nodes if node.is_nested_func
            ]

        if self.config.ignore_nested_classes:
            filtered_nodes = self._filter_inner_nested(filtered_nodes)

        if self.config.docstring_style == "google":
            self._set_google_style(filtered_nodes)

        return filtered_nodes

    def get_coverage(
        self,
    ) -> tuple[dict[str, list[CovNode]], dict[str, list[CovNode]]]:
        filenames = self.get_filenames_from_path()
        return self._get_coverage(filenames)



import ast
import os.path
from typing import Self
import attr
from genpydoc.config.config import Config

DocumentableFunc = ast.AsyncFunctionDef | ast.FunctionDef
DocumentableFuncOrClass = DocumentableFunc | ast.ClassDef
DocumentableNode = DocumentableFuncOrClass | ast.Module


@attr.s(eq=False)
class CovNode:
    """Coverage of an AST Node.

    Args:
        str name: Name of node (module, class, method or function
            names).
        str path: Pseudo-import path to node (i.e. ``sample.py:
            MyClass.my_method``).
        int level: Level of recursiveness/indentation.
        int lineno: Line number of class, method, or function.
        bool covered: Has a docstring.
        str node_type: Type of node (e.g "module", "class", or
            "function").
        bool is_nested_func: If the node itself is a nested function
            or method.
        bool is_nested_cls: If the node itself is a nested class.
        str docstring: The docstring of the node.
        str code: The source code of the node.
        CovNode parent: Parent node of current CovNode, if any.
    """

    name: str = attr.field()
    path: str = attr.field()
    level: int = attr.field()
    lineno: int = attr.field()
    covered: bool = attr.field()
    node_type: str = attr.field()
    is_nested_func: bool = attr.field()
    is_nested_cls: bool = attr.field()
    parent: Self = attr.field()
    file: str = attr.field()
    docstring: str | None = attr.field(default=None)
    code: str | None = attr.field(default=None)


class Visitor(ast.NodeVisitor):
    """Visitor is a NodeVisitor that traverses a Python AST to collect information about
    docstrings for modules, classes, and functions. During traversal, it builds
    CovNode records for each visitable node, including metadata such as the node's
    name, path, type, location, and associated docstring (when present). It also
    provides sanitized code with any docstrings removed and applies configuration
    driven ignore rules (e.g., private/semiprivate, __init__, magic methods,
    property decorators, overloads).

    Attributes:
        filename (str): The name of the file being analyzed.
        config (Config): Configuration controlling ignore rules and options.
        source (str): The full source code corresponding to the AST being visited.
        stack (list[CovNode]): Stack of CovNodes representing the current nesting context.
        nodes (list[CovNode]): All CovNodes discovered during traversal.
    """

    def __init__(self, filename: str, config: Config, source: str):
        self.filename = filename
        self.config = config
        self.source: str = source
        self.stack: list[CovNode] = []
        self.nodes: list[CovNode] = []

    @staticmethod
    def _has_doc(node: DocumentableNode) -> bool:
        """Return whether the node has docstrings."""
        return (
            ast.get_docstring(node) is not None
            and ast.get_docstring(node).strip() != ""
        )

    @staticmethod
    def _get_sanitized_docstring(node: DocumentableNode) -> str:
        """Returns a sanitized (stripped of whitespace) version of the docstring."""
        return ast.get_docstring(node).strip()

    @staticmethod
    def _remove_docstring_from_source(code: str, docstring: str) -> str:
        """Removes docstrings from the source code."""
        docstring = ['"""', *docstring.splitlines()]
        return "\n".join(
            (
                line
                for line in code.splitlines()
                if line.strip() not in docstring
            )
        )

    def _get_sanitized_code(self, node: DocumentableNode) -> str | None:
        """Returns a code segment for a node, sanitized of any docstrings."""
        code = ast.get_source_segment(self.source, node)
        if self._has_doc(node) and code:
            code = self._remove_docstring_from_source(
                code=code, docstring=self._get_sanitized_docstring(node)
            )
        return code

    def _visit_helper(self, node: DocumentableNode) -> None:
        """Recursively visit AST node for docstrings."""
        file = os.path.basename(self.filename)
        if not hasattr(node, "name"):
            node_name = os.path.basename(self.filename)
        else:
            node_name = node.name
        parent = None
        path = node_name
        if self.stack:
            parent = self.stack[-1]
            parent_path: str = parent.path
            path = (":" if parent_path.endswith(".py") else ".").join(
                [parent_path, node_name]
            )
        lineno = None
        if hasattr(node, "lineno"):
            lineno = node.lineno
        node_type = type(node).__name__
        cov_node = CovNode(
            name=node_name,
            path=path,
            covered=self._has_doc(node),
            level=len(self.stack),
            node_type=node_type,
            lineno=lineno,
            is_nested_func=self._is_nested_func(parent, node_type),
            is_nested_cls=self._is_nested_cls(parent, node_type),
            parent=parent,
            file=file,
            docstring=(
                self._get_sanitized_docstring(node)
                if self._has_doc(node)
                else None
            ),
            code=self._get_sanitized_code(node),
        )
        self.stack.append(cov_node)
        self.nodes.append(cov_node)
        self.generic_visit(node)
        self.stack.pop()

    @staticmethod
    def _is_nested_func(parent: CovNode | None, node_type: str) -> bool:
        """Is node a nested func/method of another func/method."""
        if parent is None:
            return False
        if parent.node_type == "FunctionDef" and node_type == "FunctionDef":
            return True
        return False

    @staticmethod
    def _is_nested_cls(parent: CovNode | None, node_type: str) -> bool:
        """Is node a nested func/method of another func/method."""
        if parent is None:
            return False
        if (
            parent.node_type in ["ClassDef", "FunctionDef"]
            and node_type == "ClassDef"
        ):
            return True
        return False

    @staticmethod
    def _is_private(node: DocumentableFuncOrClass) -> bool:
        """Is node private (i.e. __MyClass, __my_func)."""
        if node.name.endswith("__"):
            return False
        if not node.name.startswith("__"):
            return False
        return True

    @staticmethod
    def _is_semiprivate(node: DocumentableFuncOrClass) -> bool:
        """Is node semiprivate (i.e. _MyClass, _my_func)."""
        if node.name.endswith("__"):
            return False
        if node.name.startswith("__"):
            return False
        if not node.name.startswith("_"):
            return False
        return True

    def _is_ignored_common(self, node: DocumentableFuncOrClass) -> bool:
        """Commonly-shared ignore checkers."""
        is_private = self._is_private(node)
        is_semiprivate = self._is_semiprivate(node)
        if self.config.ignore_private and is_private:
            return True
        if self.config.ignore_semiprivate and is_semiprivate:
            return True
        return False

    def _is_class_ignored(self, node: DocumentableFuncOrClass) -> bool:
        """Should the AST visitor ignore this class node."""
        return self._is_ignored_common(node)

    def _is_func_ignored(self, node: DocumentableFuncOrClass) -> bool:
        """Should the AST visitor ignore this func/method node."""
        is_init = node.name == "__init__"
        is_magic = all(
            [
                node.name != "__init__",
                node.name.startswith("__"),
                node.name.endswith("__"),
            ]
        )
        has_property_decorators = self._has_property_decorators(node)
        has_setters = self._has_setters(node)
        has_overload = self._has_overload(node)
        if self.config.ignore_init_method and is_init:
            return True
        if self.config.ignore_magic and is_magic:
            return True
        if self.config.ignore_property_decorators and has_property_decorators:
            return True
        if self.config.ignore_property_setters and has_setters:
            return True
        if self.config.ignore_overloaded_functions and has_overload:
            return True
        return self._is_ignored_common(node)

    @staticmethod
    def _has_property_decorators(node: DocumentableFuncOrClass) -> bool:
        """Detect if node has property get/setter/deleter decorators."""
        if not hasattr(node, "decorator_list"):
            return False
        else:
            for dec in node.decorator_list:
                if hasattr(dec, "id"):
                    if dec.id == "property":
                        return True
                if hasattr(dec, "attr"):
                    if dec.attr == "setter":
                        return True
                    if dec.attr == "deleter":
                        return True
            return False

    @staticmethod
    def _has_setters(node: DocumentableFuncOrClass) -> bool:
        """Detect if node has property get/setter decorators."""
        if not hasattr(node, "decorator_list"):
            return False
        else:
            for dec in node.decorator_list:
                if hasattr(dec, "attr"):
                    if dec.attr == "setter":
                        return True
            return False

    @staticmethod
    def _has_overload(node: DocumentableFuncOrClass) -> bool:
        """Detect if node has a typing.overload decorator."""
        if not hasattr(node, "decorator_list"):
            return False
        else:
            for dec in node.decorator_list:
                if (
                    hasattr(dec, "attr")
                    and hasattr(dec, "value")
                    and hasattr(dec.value, "id")
                    and dec.value.id
                    and (dec.attr == "overload")
                ):
                    return True
                if hasattr(dec, "id") and dec.id == "overload":
                    return True
            return False

    def visit_Module(self, node: DocumentableNode) -> None:
        """Visit module for docstrings.

        Args:
            node (ast.Module): a module AST node.
        """
        self._visit_helper(node=node)

    def visit_ClassDef(self, node: DocumentableFuncOrClass) -> None:
        """Visit class for docstrings.

        Args:
            node (ast.ClassDef): a class AST node.
        """
        if self._is_class_ignored(node):
            return
        self._visit_helper(node=node)

    def visit_FunctionDef(self, node: DocumentableFuncOrClass) -> None:
        """Visit function or method for docstrings.

        Args:
            node (ast.FunctionDef): a function/method AST node.
        """
        if self._is_func_ignored(node):
            return
        self._visit_helper(node=node)

    def visit_AsyncFunctionDef(self, node: DocumentableFuncOrClass) -> None:
        """Visit async function or method for docstrings.

        Args:
            node (ast.AsyncFunctionDef): an async function/method AST
                node.
        """
        if self._is_func_ignored(node):
            return
        self._visit_helper(node=node)



from os.path import commonprefix
from pathlib import Path
import os
import tomllib
from typing import Any, Sequence

import click
from click import Context, Parameter


def find_project_root(srcs: list[str]) -> Path:
    if not srcs:
        return Path("/").resolve()
    common_base = min(Path(src).resolve() for src in srcs)
    if common_base.is_dir():
        common_base /= "_"

    for directory in common_base.parents:
        if (directory / ".git").exists():
            return directory
        if (directory / "pyproject.toml").is_file():
            return directory

    return directory


def parse_pyproject_toml(path_config: str) -> dict[str, Any] | None:
    with open(path_config, "rb") as file:
        toml = tomllib.load(file)
    config = toml.get("tool", {}).get("genpydoc")
    return {k.replace("-", "_"): v for k, v in config.items()}


def read_config_file(
    ctx: Context, _param: Parameter, value: str | None
) -> str | None:
    if not value:
        paths = ctx.params.get("paths")
        if not paths:
            paths = (os.path.abspath(os.getcwd()),)
        value = find_project_config(paths)
        if value is None:
            return None

    print(f"{value=}")

    if value.suffix == ".toml":
        try:
            config = parse_pyproject_toml(value)
        except (tomllib.TOMLDecodeError, OSError) as err:
            raise click.FileError(
                filename=value,
                hint=f"Error reading configuration file: {err}.",
            )

    if ctx.default_map is None:
        ctx.default_map = {}

    ctx.default_map.update(config)
    return value


def find_project_config(path_search_start: Sequence[str]) -> str | None:
    """Find the absolute filepath to a pyproject.toml if it exists."""
    project_root = find_project_root(path_search_start)
    pyproject_toml = project_root / "pyproject.toml"
    if pyproject_toml.is_file():
        return str(pyproject_toml)
    return None


def get_common_base(files: list[str]) -> str:
    commonbase = Path(commonprefix(files))
    while not commonbase.exists():
        commonbase = commonbase.parent
    return str(commonbase)



import os.path
import sys
from pathlib import Path

from git import Diff, Repo
from genpydoc.git_retriever.utils import process_git_diff
from genpydoc.extractor.visit import CovNode


class GitRetriever:
    def __init__(
        self,
        root: str | Path,
        covered_nodes: dict[str, list[CovNode]],
        nodes: dict[str, list[CovNode]],
    ):
        self.root = root
        self.repo = Repo(root)
        self.covered_nodes = covered_nodes
        self.nodes = nodes
        self.lines = {}
        self.__add_all()
        self._diffed_map = self.__build_diffed_map()
        if not self._diffed_map or all(
            (
                k not in self.covered_nodes.keys()
                for k in self._diffed_map.keys()
            )
        ):
            self.__stop_early()

    def __add_all(self) -> None:
        self.repo.git.add(all=True)

    def __build_diffed_map(self) -> dict[str, str]:
        def _reverse_mapping(ct: str | None) -> str:
            mapping = {"D": "A", "A": "D"}
            if ct not in mapping:
                return ct
            return mapping[ct]

        d = self.repo.index.diff("HEAD")
        return {
            os.path.join(self.root, c.a_path): _reverse_mapping(c.change_type)
            for c in d
        }

    @staticmethod
    def __stop_early() -> None:
        """Ends the program early"""
        sys.exit()

    @staticmethod
    def _process_diff(diff: Diff) -> set[int]:
        return process_git_diff(diff)

    def _extract_lines(self) -> dict[str, set[CovNode]]:
        lines_for_evaluation: dict[str, set[CovNode]] = {}
        for k in self._diffed_map.keys():
            diff = self.repo.index.diff("HEAD", paths=k, create_patch=True)
            if len(diff) > 1:
                raise ValueError
            if len(diff):
                diff = diff[0]
            if self._diffed_map.get(k, "A") == "A" and k in self.nodes:
                lines_for_evaluation[k] = self.nodes[k]
            else:
                lines = self._match_lines_to_ast(k, self._process_diff(diff))
                lines_for_evaluation[k] = lines
        return lines_for_evaluation

    def _match_lines_to_ast(self, k: str, lines: set[int]) -> set[CovNode]:
        definitions = set()
        for line in lines:
            traversed_nodes: list[CovNode] = []
            if k in self.nodes:
                for node in self.nodes[k]:
                    if node.level == 0:
                        continue
                    if (
                        node.lineno
                        <= line
                        < node.lineno + len(node.code.splitlines())
                    ):
                        traversed_nodes.append(node)
            if len(traversed_nodes) > 0:
                for n in traversed_nodes:
                    definitions.add(n)
            else:
                continue
        return definitions

    @staticmethod
    def _analyze_covered_nodes(
        diffed_nodes: dict[str, set[CovNode]],
    ) -> dict[str, set[CovNode]]:
        keys = list(diffed_nodes.keys())
        for k in keys:
            nodes = {node for node in diffed_nodes[k] if node.covered}
            if not nodes:
                del diffed_nodes[k]
                continue
            diffed_nodes[k] = nodes
        return diffed_nodes

    def extract_diff(self) -> dict[str, set[CovNode]]:
        nodes_diffed = self._extract_lines()
        return self._analyze_covered_nodes(nodes_diffed)



import os

import click
from genpydoc.config.config import Config
from genpydoc.documenter import Documenter
from genpydoc.utils.utils import read_config_file


@click.command()
@click.option(
    "-m",
    "--ignore-magic",
    is_flag=True,
    default=False,
    show_default=False,
    help=(
        "Ignore all magic methods of classes.  [default: False]\n\nNOTE: This "
        "does not include the `__init__` method. To ignore `__init__` methods, "
        "use `--ignore-init-method`."
    ),
)
@click.option(
    "-C",
    "--ignore-nested-classes",
    is_flag=True,
    default=False,
    show_default=True,
    help="Ignore nested classes.",
)
@click.option(
    "-n",
    "--ignore-nested-functions",
    is_flag=True,
    default=False,
    show_default=True,
    help="Ignore nested functions and methods.",
)
@click.option(
    "-O",
    "--ignore-overloaded-functions",
    is_flag=True,
    default=False,
    show_default=True,
    help="Ignore `@typing.overload`-decorated functions.",
)
@click.option(
    "-p",
    "--ignore-private",
    is_flag=True,
    default=False,
    show_default=False,
    help=(
        "Ignore private classes, methods, and functions starting with two "
        "underscores.  [default: False]"
        "\n\nNOTE: This does not include magic methods; use `--ignore-magic` "
        "and/or `--ignore-init-method` instead."
    ),
)
@click.option(
    "-P",
    "--ignore-property-decorators",
    is_flag=True,
    default=False,
    show_default=True,
    help="Ignore methods with property setter/getter/deleter decorators.",
)
@click.option(
    "-S",
    "--ignore-setters",
    is_flag=True,
    default=False,
    show_default=True,
    help="Ignore methods with property setter decorators.",
)
@click.option(
    "-s",
    "--ignore-semiprivate",
    is_flag=True,
    default=False,
    show_default=True,
    help=(
        "Ignore semiprivate classes, methods, and functions starting with a "
        "single underscore."
    ),
)
@click.option(
    "-o",
    "--include-only-covered",
    is_flag=True,
    default=False,
    show_default=True,
    help="Only include Node that have a docstring in the processing.",
)
@click.option(
    "-D",
    "--run-on-diff",
    is_flag=True,
    default=False,
    show_default=True,
    help="Only run the evaluator on Git diffed Nodes.",
)
@click.option(
    "--use-llm-provider",
    type=click.Choice(["openai"]),
    default="openai",
    show_default=True,
    help="Select the LLM provider.",
)
@click.option(
    "--use-model",
    type=click.Choice(["gpt-5-nano"]),
    default="gpt-5-nano",
    show_default=True,
    help="Select which LLM model to use for documenting.",
)
@click.option(
    "--style",
    type=click.Choice(["google", "numpy", "epytext", "reST"]),
    default="google",
    show_default=True,
    help="Docstring types allowed.",
)
@click.help_option("-h", "--help")
@click.argument(
    "paths",
    type=click.Path(
        exists=True,
        file_okay=True,
        dir_okay=True,
        writable=False,
        readable=True,
        resolve_path=True,
    ),
    is_eager=True,
    nargs=-1,
)
@click.option(
    "-c",
    "--config",
    type=click.Path(
        exists=False, file_okay=True, dir_okay=False, readable=True
    ),
    is_eager=True,
    callback=read_config_file,
    help="Read configuration from `pyproject.toml` or `setup.cfg`.",
)
def main(
    paths: list[str] | None,
    ignore_magic: bool,
    ignore_private: bool,
    ignore_semiprivate: bool,
    ignore_nested_classes: bool,
    ignore_nested_functions: bool,
    ignore_setters: bool,
    ignore_property_decorators: bool,
    ignore_overloaded_functions: bool,
    include_only_covered: bool,
    run_on_diff: bool,
    use_llm_provider: str,
    use_model: str,
    style: str,
    config: str | None,
):
    config = Config(
        docstring_style=style,
        ignore_magic=ignore_magic,
        ignore_private=ignore_private,
        ignore_semiprivate=ignore_semiprivate,
        ignore_nested_classes=ignore_nested_classes,
        ignore_nested_functions=ignore_nested_functions,
        ignore_property_setters=ignore_setters,
        ignore_property_decorators=ignore_property_decorators,
        ignore_overloaded_functions=ignore_overloaded_functions,
        include_only_covered=include_only_covered,
        run_on_diff=run_on_diff,
        use_llm_provider=use_llm_provider,
        use_model=use_model,
    )
    if not paths:
        paths = [os.path.abspath(os.getcwd())]

    doc = Documenter(config)
    doc.document(paths)
